const {
  saveGame,
  logError,
  getCPS,
  squadUpgrades,
  getItemCps,
  squadIsMissing,
  maybeNews,
  idFromWord,
  getCoins,
  getUser,
  commas,
  addAchievement,
  game: {nfts, squad, users}
} = require('./utils')
const slack = require('../../slack')
const buyableItems = require('./buyableItems')
const upgrades = require('./upgrades')
const achievements = require('./achievements')
const webapi = require('./webapi')
const prestige = require('./prestige')

// const readline = require('readline').createInterface({
//   input: process.stdin,
//   output: process.stdout
// })
// const read = () => {
//   readline.question(`What do YOU want? `, async want => {
//     want && await slack.messageSage(want)
//     read()
//   })
// }
// ;(async () => {
//   read()
// })();

const getUpgradeEmoji = upgrade => upgrade.emoji || buyableItems[upgrade.type].emoji
const upgradeText = (user, showOwned = false) => {
  const userDoesNotHave = ([upgradeName, upgrade]) => hasUpgrade(user, upgrade, upgradeName) === showOwned
  const userMeetsCondition = ([, upgrade]) => upgrade.condition(user)
  const format = ([key, value]) => `:${getUpgradeEmoji(value)}: *${key}* - ${commas(value.cost)}\n_${value.description}_`
  return '\n\n' +
    Object.entries(upgrades)
    .filter(userDoesNotHave)
    .filter(userMeetsCondition)
    .map(format)
    .join('\n\n') +
    '\n\n:grey_question::grey_question::grey_question:' +
    '\n\nJust type \'!upgrade upgrade_name\' to purchase'
}

const hasUpgrade = (user, upgrade, upgradeName) => !!user.upgrades[upgrade.type]?.includes(upgradeName)

const alwaysAccessible = () => true
const adminOnly = {
  hidden: true,
  condition: ({ event, say }) => {
    if (!event.user.startsWith(slack.sageUserId)) {
      say(`This is an admin-only command!`)
      return false
    }
    return true
  }
}
const testOnly = {
  hidden: true,
  condition: ({ event }) => event.user.includes('TEST')
}
const dmsOnly = {
  hidden: false,
  condition: ({ event, say, words }) => {
    if (event.channel_type !== 'im') {
      say(`Please use ${words[0]} in DMs only!`)
      return false
    }
    return true
  }
}
const prestigeOnly = adminOnly // TODO ({ event }) => !!getUser(event.user).prestige

const commands = new Map()
let commandHelpText = ''
const defaultAccess = { hidden: false, condition: alwaysAccessible }
const command = (commandNames, helpText, action, { hidden, condition } = defaultAccess) => {
  if (!hidden) {
    commandHelpText += `\n${commandNames.toString().replace(/,/g, ', ')} - ${helpText}\n`
  }
  if (!condition) {
    condition = alwaysAccessible
  }
  if (Array.isArray(condition)) {
    const conditionList = condition
    condition = arg => conditionList.every(c => c(arg))
  }
  const c = {
    commandNames,
    helpText,
    action,
    condition,
    hidden
  }
  webapi.addCommand(c)
  commandNames.forEach(name => {
    if (commands.get(name)) {
      throw `Duplicate command '${name}' detected.`
    }
    commands.set(name, c)
  })
}

const messageHandler = async ({ event, say }) => {
  const words = event?.text?.split(/\s+/) || []
  const c = commands.get(words[0])
  if (!c && words[0]?.startsWith('!')) {
    return slack.messageSage(`${slack.ourUsers[event.user]} tried to use \`${event.text}\`, if you wanted to add that.`)
  }
  const canUse = await c?.condition({ event, say, words })
  if (!canUse) {
    // await say(`Command '${words[0]}' not found`)
    return
  }
  if (words[1] === 'help') {
    await say(c.commandNames.map(name => '`' + name + '`').join(', ') + ': ' + c.helpText)
    if (c.commandNames.includes('!coin')) {
      addAchievement(getUser(event.user), 'weAllNeedHelp', say)
    }
    return
  }
  await c.action({ event, say, words })
}

slack.onMessage(async msg => {
  try {
    await messageHandler(msg)
  } catch (e) {
    logError(e)
  }
})

command(
  ['!cleanusers'],
  'Calls getUser() on all users, ensuring a valid state.',
  async ({ say }) => {
    Object.keys(users).forEach(userId => { getUser(userId) })
    return say('```Cleaning ' + JSON.stringify(Object.keys(users), null, 2) + '```')
  }, adminOnly)

command(
  ['!setpw'],
  'Set your api password. May not contain spaces. *This is not secure!*\n' +
    '  To use, say !setpw your_password',
  async ({ event, say, words }) => {
    const user = getUser(event.user)
    user.pwHash = webapi.makeHash(words[1])
    await saveGame()
    await say(`Password encoded as ${user.pwHash}`)
  }
  , { hidden: true })

command(
  ['!help', '!h'],
  'List available commands',
  async ({ say }) => say('```' + commandHelpText + '```')
)

const removeAchievement = async (user, name, say) => {
  if (user.achievements[name]) {
    user.achievements[name] = false
    await say('Achievement removed!')
  } else {
    await say('That user doesn\'t have that achievement!')
  }
}

command(
  ['!rach'],
  'Remove achievement',
  async ({ say, words }) => {
    const achName = words[1]
    const target = idFromWord(words[2])
    await removeAchievement(getUser(target), achName, say)
  }, adminOnly)

command(
  ['!a', '!ach', '!achievements'],
  'List your glorious achievements',
  async ({ event, say }) => {
    const user = getUser(event.user)

    const achievementCount = Object.keys(user.achievements).length
    const prefix = `You have ${achievementCount} achievements!\n\n`
    const mult = (Math.pow(1.01, achievementCount) - 1) * 100
    let list = ''
    if (event.channel_type === 'im') {
      list = Object.keys(user.achievements)
        .map(name => achievements[name])
        .map(({ description, emoji, name }) => `:${emoji}: *${name}* - ${description}`)
        .join('\n') + '\n\n'
    }

    const postfix = achievementCount ? `_Achievements are boosting your CPS by ${mult.toPrecision(3)}%_` : ''
    await say(prefix + list + postfix)
  }
)

const emojiLine2 = (itemName, countOwned) =>
  ((emoji = buyableItems[itemName].emoji) => countOwned < 5
    ? `:${emoji}:`.repeat(countOwned)
    : `:${emoji}: x${countOwned}`)()

const emojiLine3 = (itemName, countOwned) =>
  (countOwned < 5
    ? s => s.repeat(countOwned)
    : s => s + ` x${countOwned}`)
  (buyableItems[itemName].emoji)

const emojiLine = (itemName, countOwned) => countOwned < 5
  ? `:${buyableItems[itemName].emoji}:`.repeat(countOwned)
  : `:${buyableItems[itemName].emoji}: x${countOwned}`
const collection = userId =>
  Object.entries(users[userId]?.items || {})
    .map(([itemName, countOwned]) => emojiLine(itemName, countOwned) + ' - ' + commas(Math.round(getItemCps(getUser(userId), itemName))) + ' cps')
    .join('\n')

command(['!cps'],
    'Display your current Coins Per Second',
    async ({ event, say }) =>
        say(`You are currently earning \`${commas(getCPS(event.user))}\` HVAC Coin per second.`))

command(
  ['!c', '!coin', '!mine', '!'],
  'Mine HVAC coins',
  async ({ event, say }) => {
    const user = getUser(event.user)
    maybeNews(say)
    const random = Math.random()
    const c = getCoins(event.user)
    const secondsOfCps = seconds => Math.floor(getCPS(event.user) * seconds)
    const hour = new Date().getHours()
    if (hour < 8 || hour > 18) {
      addAchievement(user, 'hvackerAfterDark', say)
    }
    let diff
    let prefix
    if (random > 0.9967) {
      diff = 500 + Math.floor(c * 0.10) + secondsOfCps(60 * 30)
      prefix = `:gem: You found a lucky gem worth ${commas(diff)} HVAC!\n`
      addAchievement(user, 'luckyGem', say)
      await slack.messageSage(`${slack.ourUsers[event.user]} FOUND A LUCKY GEM COIN WORTH ${commas(diff)} HVAC!`)
    } else if (random > 0.986) {
      diff = 50 + Math.floor(c * 0.025) + secondsOfCps(60)
      prefix = `:goldbrick: You found a lucky gold coin worth ${commas(diff)} HVAC!\n`
      addAchievement(user, 'goldBrick', say)
    } else if (random > 0.96) {
      diff = 10 + Math.floor(c * 0.01) + secondsOfCps(10)
      prefix = `:money_with_wings: You found a lucky green coin worth ${commas(diff)} HVAC!\n`
      addAchievement(user, 'greenCoin', say)
    } else {
      prefix = 'You mined one HVAC.\n'
      diff = 1
    }
    user.coins += diff
    await say(`${prefix}You now have ${commas(user.coins)} HVAC coin` + (c !== 1 ? 's' : '') + '. Spend wisely.')
    saveGame()
  }
)

command(
  ['!g', '!gamble'],
  'Gamble away your HVAC\n' +
    '  To use, say \'gamble coin_amount\' or \'!gamble all\'',
  async ({ event, say, words }) => {
    getCoins(event.user) // Update coin counts
    const user = getUser(event.user)

    let n
    if (words[1].toLowerCase() === 'all') {
      if (user.coins === 0) {
        return say('You don\'t have any coins!')
      }
      n = user.coins
    } else {
      n = parseInt(words[1])
    }
    if (!n || n < 0) {
      return say(`Invalid number '${n}'`)
    }
    if (user.coins < n) {
      return say(`You don\'t have that many coins! You have ${commas(user.coins)}.`)
    }
    if (n >= 100_000_000_000) {
      addAchievement(user, 'bigBets', say)
    }
    user.coins -= n
    let outcome
    if (Math.random() > 0.5) {
      user.coins += (2 * n)
      outcome = 'won'
    } else {
      outcome = 'lost'
    }
    console.log(`They ${outcome}`)
    await say(`You bet ${commas(n)} coins and ${outcome}! You now have ${commas(user.coins)}.`)
    saveGame()
  }
)

command(
  ['!buynft', '!bn'],
  'Acquire high-quality art\n' +
    '  To use, say \'!buynft nft_name\'',
  async ({ event, say, words }) => {
    const nft = nfts.find(n => n.name.toLowerCase() === words[1])
    if (!nft) {
      const suffix = words[1]?.match(/[^a-z0-9_]/i) ? '. And I\'m unhackable, so cut it out.' : ''
      return say('No NFT with that name found' + suffix)
    }
    if (nft.owner) {
      return say('Someone already owns that NFT!')
    }
    const c = getCoins(event.user)
    if (c < nft.price) {
      return say('You don\'t have enough coin for this nft')
    }

    users[event.user].coins -= nft.price
    nft.owner = event.user
    saveGame()
    await say('You bought ' + nft.name + '!')
  }
)

command(
  ['!myupgrades', '!myu'],
  'List all the upgrades that you own.',
  async ({ event, say }) => {
    const user = getUser(event.user)
    await say(upgradeText(user, true))
  }, dmsOnly)

command(
  ['!upgrade', '!u'],
  'Improve the performance of your HVAC-generators.\n' +
    '  Say \'!upgrade\' to list available upgrades, or \'!upgrade upgrade_name\' to purchase.',
  async ({ event, say, words }) => {
    const user = getUser(event.user)
    const upgradeName = words[1]
    if (!upgradeName) {
      return say(upgradeText(user))
    }
    const upgrade = upgrades[upgradeName]
    if (!upgrade) {
      return say('An upgrade with that name does not exist!')
    }
    if (!user.upgrades[upgrade.type]) {
      user.upgrades[upgrade.type] = []
    }
    if (hasUpgrade(user, upgrade, upgradeName)) {
      return say('You already have that upgrade!')
    }
    if (!upgrade.condition(user)) {
      return say('That item does not exist!')
    }
    const c = getCoins(event.user)
    if (c < upgrade.cost) {
      return say(`You don't have enough coins! You have ${commas(c)}, but you need ${commas(upgrade.cost)}`)
    }
    user.coins -= upgrade.cost
    user.upgrades[upgrade.type].push(upgradeName)
    await saveGame()
    await say(`You bought ${upgradeName}!`)
  }, dmsOnly)

const getCurrentSquadgrade = () => {
  const current = Object.entries(squadUpgrades).find(squadIsMissing)
  if (!current) {
    return current
  }
  const [name, upgrade] = current
  if (!squad.upgrades[name]) {
    squad.upgrades[name] = upgrade.cost
  }
  return {
    name,
    remaining: squad.upgrades[name],
    emoji: upgrade.emoji,
    description: upgrade.description
  }
}

const squadgradeText = ([name, { emoji, description }]) =>
    `:${emoji}: *${name}*\n_${description}_`

const squadText = () => {
  const current = getCurrentSquadgrade()
  if (current) {
    const currentUpgradeText = ({ name, remaining, emoji, description }) =>
        `:${emoji}: *${name}* - ${commas(remaining)} HVAC remaining.\n_${description}_`
    return currentUpgradeText(current)
  }
  return 'No more squadgrades currently available.'
  // const squadIsMissing = ([name, upgrade]) => squad.upgrades[name] === false
  // return squadgradeText(Object.entries(squadUpgrades).find(squadIsMissing))
}

command(
  ['!squad', '!sq'],
  'Buy upgrades that help the whole team.\n' +
    '  Say !squad to list squad upgrades.\n' +
    '  Say \'!squad contrib_amount\' to make progress toward the current upgrade.',
  async ({ event, say, words }) => {
    if (!words[1]) {
      return say(squadText())
    }
    const current = getCurrentSquadgrade()
    if (!current) {
      return say('No squadgrades are currently available')
    }
    const currentCoins = getCoins(event.user)
    let amount
    if (words[1] === 'all') {
      amount = currentCoins
    } else {
      amount = parseInt(words[1])
      if (amount > currentCoins) {
        return say(`You don't have that much HVAC! You have ${currentCoins}.`)
      }
    }
    if (!amount || amount < 1) {
      return say(`Invalid amount: '${words[1]}'`)
    }
    if (amount > squad.upgrades[current.name]) {
      amount = squad.upgrades[current.name]
    }
    const user = getUser(event.user)
    squad.upgrades[current.name] -= amount
    user.coins -= amount
    if (squad.upgrades[current.name] <= 0) {
      squad.upgrades[current.name] = true
    }
    saveGame()
    await say(`Thank you for your contribution of ${commas(amount)} HVAC! Current status:\n\n${squadText()}`)
  }
)

const buyRoute = require('./buy')
command(
    ['!buy', '!b'],
    'Buy new items to earn HVAC with\n' +
    '  Use without arguments to list all available items.\n' +
    '  Say \'!buy item_name optional_quantity\' to make your purchase.',
    buyRoute
)

command(
  ['!check', '!ch'],
  'Check how many coins another player has',
  async ({ say, words }) => {
    const targetId = idFromWord(words[1])
    if (!targetId) {
      return say('Target must be a valid @')
    }

    if (targetId !== 'U0344TFA7HQ') {
      const coins = getCoins(targetId)
      await say(`<@${targetId}> has ${commas(coins)} HVAC.`)
    } else {
      const members = (await slack.app.client.conversations.members({ channel: slack.temperatureChannelId })).members
      const humanMembers = members.filter(name => name.length === 11)
      await say(`Hvacker owns ${humanMembers.length} souls.`)
    }
  }
)

command(
  ['!gift', '!give', '!gi'],
  'Donate coins to a fellow player\n' +
    '  Send coins by saying \'!gift @player coin_amount\'',
  async ({ event, words, say }) => {
    const userId = event.user
    const user = getUser(userId)
    const [, target, amountText] = words
    const amount = amountText === 'all' ? getCoins(userId) : parseInt(amountText)
    const targetId = idFromWord(target)
    if (!amount || amount < 0) {
      return say('Amount must be a positive integer!')
    }
    if (!targetId) {
      return say('Target must be a valid @')
    }
    if (user.coins < amount) {
      return say(`You don't have that many coins! You have ${commas(user.coins)} HVAC.`)
    }
    if (amountText === 'all' && slack.ourUsers[targetId] === 'Tyler') {
      addAchievement(user, 'walmartGiftCard', say)
    }
    const targetUser = getUser(targetId)
    user.coins -= amount
    targetUser.coins += amount
    await say(`Gifted ${commas(amount)} HVAC to <@${targetId}>`)
  }
)

command(
  ['!status', '!s'],
  'Print your current CPS, HVAC balance, and owned items',
  async ({ event, say }) => {
    await say(
          `You are currently earning \`${commas(getCPS(event.user))}\` HVAC Coin per second.\n\n` +
          `You currently have ${commas(getCoins(event.user))} HVAC Coins\n\n` +
          `${collection(event.user)}\n\nCoins collected all-time: ${commas(Math.round(users[event.user].coinsAllTime))}\n\n`
    )
  }
)

command(
  ['!gimme'],
  'Give self x coins',
  async ({ event, say, words }) => {
    const user = getUser(event.user)
    const increase = parseInt(words[1].replace(/,/g, ''))
    user.coins += increase
    user.coinsAllTime += increase
    await say(`You now have ${getCoins(event.user)} HVAC.`)
  }, testOnly)

command(
  ['!nfts', '!nft', '!n'],
  'Show NFTs in the museum\n' +
    '  Call with no arguments to list all NFTs, or \'!nft nft_name\' to show a particular one.',
  async ({ say, words }) => {
    const owner = nft => `Owner: *${slack.ourUsers[nft.owner] || 'NONE'}*`
    const nftDisplay = nft => `_"${nft.name}"_\n\n${nft.description}\n\n${commas(nft.price)} HVAC.\n\n${nft.picture}\n\n${owner(nft)}`

    const filter = words[1] ? nft => words[1]?.toLowerCase() === nft.name : null

    await say(nfts
      .filter(filter || (() => true))
      .map(nftDisplay)
      .join('\n-------------------------\n') || (filter ? 'No NFTs with that name exist' : 'No NFTs currently exist.')
    )
  }
)

command(
  ['!leaderboard', '!lb'],
  'Show the top HVAC-earners, ranked by CPS',
  async ({ event, say }) => {
    const user = getUser(event.user)
    await say('```' +
      Object.entries(users)
        .filter(([id]) => getCPS(id) !== 0)
        .sort(([id], [id2]) => getCPS(id) > getCPS(id2) ? -1 : 1)
        .map(([id]) => `${slack.ourUsers[id] || '???'} - ${commas(getCPS(id))} CPS - ${commas(getCoins(id))} HVAC`)
        .join('\n') + '```'
    ).then(() => addAchievement(user, 'leaderBoardViewer', say))
  }
)

const oneShot = (name, helpText, message, achievementName) =>
    command([name], helpText, async ({ event, say }) => {
      await say(message)
      await slack.messageSage(`Wow buddy they like your ${name} joke.`)
      if (achievementName) {
        addAchievement(getUser(event.user), achievementName, say)
      }
    }, { hidden: true })

oneShot('!santa', 'Ho ho ho!', '<https://i.imgur.com/dBWgFfQ.png|I\'m Santa Quacks>')
oneShot('!sugma', 'GRRRR', ':hvacker_angery:')
oneShot('!pog', 'Displays a poggers hvacker', '<https://i.imgur.com/XCg7WDz.png|poggers>')
oneShot('!ligma', 'Toy with me not, son.', '<https://i.imgur.com/i1YtW7m.png|no>')
oneShot('!dab', 'ACTIVATE COOL GUY MODE', '<https://i.imgur.com/FKYdeqo.jpg|I go XD style>', 'certifiedCoolGuy')
oneShot('!based', 'Sorry, it\'s a little hard to hear you!', '<https://i.imgur.com/IUX6R26.png|What?>')
oneShot('!shrek', 'Love and life, baby.', '<https://i.imgur.com/QwuCQZA.png|Donkey!>')

command(
  ['!addnft'],
  'Arguments 1 and 2 should be on the first line as name (one word!) and integer price.\n' +
    '  The second line should be the description of the pieces\n' +
    '  the picture is everything after the second line',
  async ({ event }) => {
    const [, name, price] = event.text.substring(0, event.text.indexOf('\n')).split(' ')
    const rest = event.text.substring(event.text.indexOf('\n') + 1)
    const desc = rest.substring(0, rest.indexOf('\n'))
    const picture = rest.substring(rest.indexOf('\n') + 1)
    const newNft = {
      name,
      price: parseInt(price.replace(/,/g, '')),
      description: desc,
      picture,
      owner: null
    }
    nfts.push(newNft)
    console.log('addedNft', newNft)
    return saveGame()
  }, adminOnly)

command(
    ['!prestige', '!p'],
    'Show your current prestige status',
    prestige.prestigeRoute,
    adminOnly
)

command(
    ['!!prestige'],
    'Confirm your prestige activation.',
    prestige.prestigeConfirmRoute,
    adminOnly
)

command(
    ['!quack', '!quackstore'],
    'Spend your prestigious quackings',
    prestige.quackStoreRoute,
    prestigeOnly
)

command(
  ['!ss'],
  'Show the status for another player: !ss @player',
  async ({ words, say }) => {
    const target = words[1]
    const targetId = idFromWord(target)
    await say(
          `${target} is currently earning \`${commas(getCPS(targetId))}\` HVAC Coin per second.\n\n` +
          `They have ${commas(getCoins(targetId))} HVAC Coins\n\n` +
          `${collection(targetId)}\n\n` +
          `${Object.entries(users[targetId]?.items || {}).reduce((total, [name, count]) => total + count, 0)} total items\n\n`
    )
  }, adminOnly)

const exec = require('child_process').exec;
command(
  ['!pl'],
  '!pl `code`',
  async ({ event, say}) => {
    const code = event.text.substring(4)
      .replace(/`/g, '')
      .replace(/&lt;/g, '<')
    console.log('PL CODE:', code)
    const child = exec(`/home/sagevaillancourt/projects/pebblisp/src/pl '${code}'`)
    let result = '```'
    let errors = ''
    child.stdout.on('data', data => result += data)
    child.stderr.on('data', data => {
      result += `\nERR:${data}\n`
      errors += `\nERR:${data}`
    })
    child.on('close', async () => {
      if (result.length > 2000) {
        result = result.substring(0, 2000) + '...\nOUTPUT EXCEEDS MAXIMUM AND WAS TRUNCATED\n' +
          'STDERR:\n' + errors.substring(0, 200)
      }
      await say(result + '```')
    })
  }, adminOnly)

command(
  ['!steal', '!sagesteal'],
  'Highly illegal',
  async ({ event, words, say }) => {
    const userId = event.user
    const user = getUser(userId)
    const [, target] = words
    const amount = getCoins(userId) / 100
    if (!amount || amount < 0) {
      return
    }
    let targetId = idFromWord(target)
    if (!targetId) {
      targetId = slack.sageUserId
    }
    if (user.coins < amount) {
      return
    }
    const targetUser = getUser(targetId)
    user.coins -= amount
    targetUser.coins += amount
    await say(`Stealing is wrong. Gave ${commas(Math.floor(amount))} of your HVAC to ${slack.ourUsers[targetId]}`)
  }, { hidden: true })

// TODO
command(
  ['!lottery'],
  'Once per day, try for a big win!',
  async ({ event, say }) => {
    const user = getUser(event.user)
    const currentDate = new Date().getDate()
    const lastLotto = user.lastLotto || currentDate - 1
    if (lastLotto === currentDate) {
      return say('Hey, only one lotto ticket per day, alright?')
    }
    user.lastLotto = currentDate
    saveGame()
  }, { hidden: true })

command(
  ['!ignite'],
  'You found me!',
  async ({ event, say }) => {
    const user = getUser(event.user)
    addAchievement(user, 'ignited', say)
  }, { hidden: true })

command(
  ['!giveach'],
  '!giveach @player ach_name',
  async ({ words, say }) => {
    const targetId = idFromWord(words[1])
    const user = getUser(targetId)
    console.log(user, words[2], say)
    addAchievement(user, words[2], say)
    saveGame()
  }, adminOnly)

command(
  ['!whois'],
  '!whois player_id',
  async ({ words, say }) => say(`<@${words[1]}>`),
  adminOnly)

command(
  ['!ngift'],
  '!ngift player_id nft_name',
  async ({ event, words, say }) => {
    const targetId = idFromWord(words[1])
    if (!targetId) {
      return say(`Please specify a valid @ target!`)
    }
    const nft = nfts.find(nft => nft.name === words[2])
    if (!nft) {
      return say(`There is not NFT named "${words[2]}"!`)
    }
    if (nft.owner !== event.user) {
    return say(`You do not own "${nft.name}"!`)
    }
    nft.owner = targetId
  }, { hidden: true })

command(
  ['!deletetest', '!dtest'],
  '!deletetest',
  async () => {
    delete users[slack.testId]
    saveGame()
  }, adminOnly)

command(
  ['!message', '!msg', '!m'],
  '!message player_id message',
  async ({ event, words, say }) => {
    const targetId = idFromWord(words[1])
    if (!targetId) {
      return say(`Please specify a valid @ target!`)
    }
    const msg = event.text.replace(/\S+ +\S+\s*/i, '')
    await slack.messageIn(targetId, msg)
    await slack.messageSage(msg)
  }, adminOnly)

webapi.launch()